#include "kcl.h"
#include <stdlib.h>

void KCL_PtrPool_init(KCL_PtrPool *pool, int sizeObj, int sizeAlloc, int extendParam)
{
	char *cp;
	if (sizeAlloc < 0)
		sizeAlloc = 16;
	if (extendParam <= 0)
		extendParam = 8;
	KCL_PtrStack_init(&pool->stack);
	pool->sizeObj = sizeObj;
	pool->sizeAlloc = sizeAlloc;
	pool->extendParam = extendParam;
	cp = malloc(KCL_AlignMalloc + sizeObj * sizeAlloc);
	if (cp == 0)
		KCL_error("KCL_PtrPool_init: out of memory");
	*((KCL_Obj *) cp) = 0;
	pool->a = cp;
	cp += KCL_AlignMalloc;
	KCL_PtrStack_pushArray(&pool->stack, cp, sizeObj, sizeAlloc);
	return;
}

void KCL_PtrPool_free(KCL_PtrPool *pool)
{
	KCL_Obj p = pool->a, q;
	while (p != 0) {
		q = *((KCL_Obj *) p);
		free(p);
		p = q;
	}
	pool->a = 0;
	return;
}

void KCL_PtrPool_extend(KCL_PtrPool *pool, int sizeAlloc)
{
	char *cp;
	if (pool->sizeAlloc >= sizeAlloc) goto fin;
	cp = malloc(KCL_AlignMalloc + pool->sizeObj * (sizeAlloc - pool->sizeAlloc));
	if (cp == 0)
		KCL_error("KCL_PtrPool_extend: out of memory");
	*((KCL_Obj *) cp) = pool->a;
	pool->a = cp;
	cp += KCL_AlignMalloc;
	KCL_PtrStack_pushArray(&pool->stack, cp, pool->sizeObj, sizeAlloc - pool->sizeAlloc);
	pool->sizeAlloc = sizeAlloc;
fin:
	return;
}

void KCL_PtrPool_extendAuto(KCL_PtrPool *pool)
{
	int add = (pool->extendParam * pool->sizeAlloc) >> 4;
	if (add <= 0)
		add = 1;
	KCL_PtrPool_extend(pool, pool->sizeAlloc + add);
	return;
}

KCL_Obj KCL_PtrPool_pull(KCL_PtrPool *pool)
{
	KCL_Obj obj = KCL_PtrStack_pull(&pool->stack);
	if (obj == 0) {
		KCL_PtrPool_extendAuto(pool);
		obj = KCL_PtrStack_pull(&pool->stack);
	}
	return obj;
}

KCL_Obj KCL_PtrPool_push(KCL_PtrPool *pool, KCL_Obj obj)
{
	KCL_PtrStack_push(&pool->stack, obj);
	return obj;
}

